//////////////////////////////////////////////////////////////////////////
//	"netspeed.cpp"

#include	"socklib.h"
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<process.h>
#include	<time.h>
#include	<conio.h>

class	NetSpeed
{
public:
	enum netspeed_enum
	{
		MODE_RECV = 0,
		MODE_SEND = 1,
		ST_UDP = 0,
		ST_TCP = 1,
		DEFAULT_PORT  = 10240,
		BUFFER_SIZE   = 32768,
		TRANSFER_SIZE = 0x100000,
		MO_WAIT_TIME = 1000,	// WaitForMultipleObjects
		SELECT_WAIT_SEC = 1,
	};

public:
	int				mode;
	int				stmode;
	char		   *host;
	int				port;
	BOOL			tcpNoDelay;
	unsigned		bufferSize;
	LONGLONG		transferSize;
	int				threadNum;
	int				taskNum;
	struct TaskWork
	{
		NetSpeed		   *ns;
		int					no;
		SockLib				socklib;
		unsigned			bufferSize;
		void			   *buffer;
		LONGLONG			lsize;
		LONGLONG			lsize2;
		unsigned			threadId;
		SockLib::SockAddr	addr;
		int					errnum;
	};
	TaskWork	*task;
	HANDLE		*htbl;

	BOOL				initCritical;
	CRITICAL_SECTION	critical;

	struct Header
	{
		unsigned		size;
		unsigned		seed;
	};

public:
	NetSpeed();

	int		Run(void);

private:
	void	PrintGuide(void);
	static unsigned __stdcall recv_thread(void * param);
	static unsigned __stdcall send_thread(void * param);
};

NetSpeed::NetSpeed()
{
	this->mode = MODE_RECV;
	this->stmode = ST_UDP;
	this->host = NULL;
	this->port = DEFAULT_PORT;
	this->tcpNoDelay = FALSE;
	this->bufferSize = BUFFER_SIZE;
	this->transferSize = TRANSFER_SIZE;
	this->threadNum = 1;
	this->taskNum = 0;
	this->task = NULL;
	this->htbl = NULL;

	this->initCritical = FALSE;
}


int		NetSpeed::Run(void)
{
	int				i;
	clock_t			clk0, clk1;
	int				retcode;

	if ( (this->task = new TaskWork[this->threadNum]) == NULL )
		return -1;
	this->taskNum = this->threadNum;
	for ( i = 0; i < this->taskNum; i++ )
	{
		this->task[i].ns = this;
		this->task[i].no = i;
		this->task[i].buffer = NULL;
		this->task[i].lsize = 0;
		this->task[i].lsize2 = (LONGLONG)-1;
		this->task[i].errnum = 0;
	}

	try
	{
		if ( (this->htbl = new HANDLE[this->threadNum]) == NULL )
			throw -1;
		for ( i = 0; i < this->taskNum; i++ )
			this->htbl[i] = NULL;

		::InitializeCriticalSection(&this->critical);
		this->initCritical = TRUE;

		for ( i = 0; i < this->taskNum; i++ )
		{
			int					socktype = SOCK_DGRAM;
			TaskWork		   *task;
			const char		   *protname;

			fprintf(stderr,"thread %d : ", i + 1);
			task = &this->task[i];
			switch ( this->stmode )
			{
				case ST_UDP: socktype = SOCK_DGRAM ; protname = "UDP"; break;
				case ST_TCP: socktype = SOCK_STREAM; protname = "TCP"; break;
			}
			if ( task->socklib.Create(socktype) < 0 )
			{
				fprintf(stderr," socklib create error\n");
				throw -1;
			}
			task->bufferSize = this->bufferSize;
			if ( this->stmode == ST_UDP )
			{
				unsigned int	optval;
				int				optsize;

				optval = 0;
				optsize = sizeof(optval);
				if ( task->socklib.GetSockOpt(SOL_SOCKET, SO_MAX_MSG_SIZE, &optval, &optsize) < 0 )
				{
					fprintf(stderr," socklib getsockopt error\n");
					throw -1;
				}
				if ( optval > 0 && this->bufferSize > optval )
				{
					task->bufferSize = optval;
					if ( i == 0 )
						fprintf(stderr," change buffer size %u : ", task->bufferSize);
				}
			}
			if ( (task->buffer = new BYTE [task->bufferSize]) == NULL )
				throw -1;
			memset( task->buffer, 0, task->bufferSize);
			if ( this->mode == MODE_SEND )
			{
				unsigned	seed;
				Header	   *head;
				BYTE	   *p;

				seed = (unsigned)clock();
				srand(seed);
				head = (Header *)task->buffer;
				head->size = task->bufferSize;
				head->seed = seed;
				for ( p = (BYTE *)(head + 1); p < (BYTE *)task->buffer + task->bufferSize; p++ )
					*p = rand();
			}

			if ( this->stmode == ST_TCP )
			{
				BOOL	flag;

				flag = this->tcpNoDelay ? 1 : 0;
				if ( task->socklib.SetSockOpt(IPPROTO_TCP, TCP_NODELAY, &flag, sizeof(flag)) < 0 )
				{
					fprintf(stderr," socklib setsockopt(tcp_nodelay) error\n");
					throw -1;
				}
			}

			switch ( this->mode )
			{
				case MODE_RECV:
					task->addr.sockaddr.sin_family = AF_INET;
					task->addr.sockaddr.sin_addr.s_addr = INADDR_ANY;
					task->addr.SetPort(this->port + i);
					fprintf(stderr," bind any (port %u) %s", port + i, protname);
					if ( task->socklib.Bind(task->addr) < 0 )
					{
						fprintf(stderr," socklib bind error\n");
						throw -1;
					}
					switch ( this->stmode )
					{
						case ST_UDP:
							break;
						case ST_TCP:
							task->socklib.Listen(1);
							break;
					}
					break;

				case MODE_SEND:
				{
					task->addr.sockaddr.sin_family = AF_INET;
					task->addr.SetAddr(this->host);
					task->addr.SetPort(this->port + i);
					fprintf(stderr," connect %s (port %u) %s", inet_ntoa(task->addr.sockaddr.sin_addr), port + i, protname);
					switch ( this->stmode )
					{
						case ST_UDP:
							break;
						case ST_TCP:
							if ( task->socklib.Connect(task->addr) < 0 )
							{
								fprintf(stderr," socklib connect error\n");
								throw -1;
							}
							break;
					}
					break;
				}
			}
			fprintf(stderr,"\n");
		}

		::EnterCriticalSection(&this->critical);
		switch ( this->mode )
		{
			case MODE_RECV:
				for ( i = 0; i < this->taskNum; i++ )
				{
					if ( (this->htbl[i] = (HANDLE)_beginthreadex( NULL, 0, recv_thread, &this->task[i], 0, &this->task[i].threadId)) == 0 )
						throw -1;
				}
				break;

			case MODE_SEND:
				for ( i = 0; i < this->taskNum; i++ )
				{
					if ( (this->htbl[i] = (HANDLE)_beginthreadex( NULL, 0, send_thread,  &this->task[i], 0, &this->task[i].threadId)) == 0 )
						throw -1;
				}
				break;
		}
		clk0 = ::clock();
		::LeaveCriticalSection(&this->critical);

		for (;;)
		{
			DWORD	result;

			result = ::WaitForMultipleObjects(this->taskNum, this->htbl, TRUE, NetSpeed::MO_WAIT_TIME);
			if ( result != WAIT_TIMEOUT || _kbhit() != 0 )
				break;
			this->PrintGuide();
		}
		clk1 = ::clock();
		for ( i = 0; i < this->taskNum; i++ )
			::CloseHandle(this->htbl[i]);

		this->PrintGuide();
		switch ( this->mode )
		{
			case MODE_RECV:
			{
				LONGLONG	size;

				size = 0;
				for ( i = 0; i < this->taskNum; i++ )
					size += this->task[i].lsize;
				fprintf(stdout, "recv %u K bytes\n", (unsigned)(size/(LONGLONG)0x400));
				break;
			}

			case MODE_SEND:
				if ( clk1 > clk0 )
				{
					LONGLONG	size;
					unsigned	clk, speed;

					clk = clk1 - clk0;
					size = 0;
					for ( i = 0; i < this->taskNum; i++ )
						size += this->task[i].lsize;
					speed = (unsigned)(size * (LONGLONG)CLOCKS_PER_SEC / (LONGLONG)clk);
					fprintf(stdout, "send %u K bytes, speed %u bytes/sec (%u clocks)\n", 
					  (unsigned)(size/(LONGLONG)0x400),
					  speed,
					  clk);
				}
				break;
		}
		retcode = 0;
	} // end of try
	catch (int code) { retcode = code; }

	if ( this->initCritical )
	{
		::DeleteCriticalSection(&this->critical);
		this->initCritical = FALSE;
	}
	if ( this->htbl != NULL )
	{
		delete [] this->htbl;
		this->htbl = NULL;
	}
	if ( this->task != NULL )
	{
		for ( i = 0; i < this->taskNum; i++ )
		{
			if ( this->task[i].buffer != NULL )
				delete [] this->task[i].buffer;
		}
		delete [] this->task;
		this->task = NULL;
	}
	return retcode;
}

void	NetSpeed::PrintGuide(void)
{
	BOOL	disp;
	int		i;

	::EnterCriticalSection(&this->critical);
	disp = FALSE;
	for ( i = 0; i < this->taskNum; i++ )
	{
		if ( this->task[i].lsize != this->task[i].lsize2 )
		{
			disp = TRUE;
			break;
		}
	}
	if ( disp )
	{
		for ( i = 0; i < this->taskNum; i++ )
		{
			fprintf(stderr, " %12u ", (unsigned)(this->task[i].lsize/(LONGLONG)0x400) );
			this->task[i].lsize2 = this->task[i].lsize;
		}
		fprintf(stderr, "\n");
	}
	::LeaveCriticalSection(&this->critical);
}


unsigned __stdcall NetSpeed::recv_thread(void * param)
{
	NetSpeed::TaskWork	*task = (NetSpeed::TaskWork *)param;
	NetSpeed			*ns = task->ns;
	SockLib::SockAddr	from;
	SockLib				msgsock;
	int					ret;

	::EnterCriticalSection(&ns->critical);
	fprintf(stderr,"start thread %d\n", task->no + 1);
	::LeaveCriticalSection(&ns->critical);

	for (;;)
	{
		switch ( ns->stmode )
		{
			case ST_UDP:
			{
				fd_set			fds;
				SOCKET			fd;
				struct timeval	tm;

				tm.tv_sec  = NetSpeed::SELECT_WAIT_SEC;
				tm.tv_usec = 0;
				fd = task->socklib.GetSocket();
				FD_ZERO(&fds);
				FD_SET(fd, &fds);
				::select(fd + 1, &fds, NULL, NULL, &tm);	// M҂
				if ( FD_ISSET(fd, &fds) )
					ret = task->socklib.RecvFrom(task->bufferSize, task->buffer, 0, from);
				else
					ret = 0;
				break;
			}

			case ST_TCP:
				if ( msgsock.IsCreate() == FALSE )
				{
					if ( task->socklib.Accept(&msgsock, from) < 0 )
						goto END;
				}
				ret = msgsock.Recv(task->bufferSize, task->buffer, 0);
				if ( ret <= 0 )
					msgsock.Close();
				break;
		}
		if ( ret == 0 && task->lsize > 0 )
			break;
		if ( ret > 0 )
		{
			NetSpeed::Header	*head;

			task->lsize += ret;
			head = (NetSpeed::Header *) task->buffer;
			if ( head->size != (unsigned)ret )
			{
				::EnterCriticalSection(&ns->critical);
				task->errnum++;
				fprintf(stderr,"recv data error %u != %u bytes\n", ret, head->size);
				::LeaveCriticalSection(&ns->critical);
			}
		}
	}

END:
	 _endthreadex(0);
	return 0;
}

unsigned __stdcall NetSpeed::send_thread(void * param)
{
	NetSpeed::TaskWork	*task = (NetSpeed::TaskWork *)param;
	NetSpeed			*ns = task->ns;
	int					ret;

	::EnterCriticalSection(&ns->critical);
	fprintf(stderr,"start thread %d\n", task->no + 1);
	::LeaveCriticalSection(&ns->critical);

	while ( task->lsize < ns->transferSize )
	{
		switch ( ns->stmode )
		{
			case ST_UDP:
				ret = task->socklib.SendTo(task->bufferSize, task->buffer,0, task->addr);
				break;

			case ST_TCP:
				ret = task->socklib.Send(task->bufferSize, task->buffer,0);
				break;
		}
		if ( ret > 0 )
		{
		//	::EnterCriticalSection(&ns->critical);
		//	fprintf(stderr,"send %u bytes\n", ret);
		//	::LeaveCriticalSection(&ns->critical);
			task->lsize += ret;
		}
	}

	 _endthreadex(0);
	return 0;
}


void	help(void)
{
	static const char *msg[] =
	{
		"usage: netspeed <option>\n",
		"   -recv              ... receive(server) mode (default)\n",
		"   -send <host/addr>  ... send(client) mode\n",
		"   -udp               ... UDP (default)\n",
		"   -tcp               ... TCP\n",
		"   -tcpnodelay        ... TCP NODELAY with -tcp\n",
		"   -port <n>          ... port number\n",
		"   -thread <n>        ... thread number\n",
		"   -size <size>       ... transfer size (<n> + G, M, K)\n",
		"   -buffer <size>     ... buffer(packet) size\n",
	};
	int				i;

	for ( i = 0; i < sizeof(msg)/sizeof(*msg); i++ )
		fputs(msg[i], stdout);
}

int		main(int argc, char **argv)
{
	int				i;
	NetSpeed		ns;

	for ( i = 1; i < argc; i++ )
	{
		const char	*s;

		s = argv[i];
		if ( s[0] == '-' )
		{	// option
			++s;
			if ( strcmpi(s, "help") == 0 || s[0] == '?' )
			{
				help();
				return 1;
			} else if ( strcmpi(s, "recv") == 0 )
			{
				ns.mode = NetSpeed::MODE_RECV;
			} else if ( strcmpi(s, "send") == 0 )
			{
				ns.mode = NetSpeed::MODE_SEND;
				if ( i + 1 < argc && argv[i+1][0] != '-' )
				{	// host
					s = argv[++i];
					ns.host = (char *)s;
				}
			} else if ( strcmpi(s, "udp") == 0 )
			{
				ns.stmode = NetSpeed::ST_UDP;
			} else if ( strcmpi(s, "tcp") == 0 )
			{
				ns.stmode = NetSpeed::ST_TCP;
			} else if ( strcmpi(s, "tcpnodelay") == 0 )
			{
				ns.tcpNoDelay = TRUE;
			} else if ( strcmpi(s, "port") == 0 && i + 1 < argc && argv[i+1][0] != '-' )
			{
				struct protoent	   *pe;

				s = argv[++i];
				if ( isdigit(s[0]) )
				{
					ns.port = atoi(s);
				} else
				{
					if ( (pe = ::getprotobyname(s)) != NULL )
						ns.port = pe->p_proto;
				}
			} else if ( strcmpi(s, "thread") == 0 && i + 1 < argc && argv[i+1][0] != '-' )
			{
				int			n;

				s = argv[++i];
				n = atoi(s);
				if ( n > 1 )
					ns.threadNum = n;
			} else if ( strcmpi(s, "size") == 0 && i + 1 < argc && argv[i+1][0] != '-' )
			{
				int			n, len;

				s = argv[++i];
				len = strlen(s);
				n = atoi(s);
				if ( n > 1 )
				{
					switch ( s[len-1] )
					{
						case 'G':	// giga
						case 'g':
							ns.transferSize = (LONGLONG)n * (LONGLONG)0x40000000;
							break;

						case 'M':	// mega
						case 'm':
							ns.transferSize = (LONGLONG)n * (LONGLONG)0x100000;
							break;

						default:
						case 'K':	// kilo
						case 'k':
							ns.transferSize = (LONGLONG)n * (LONGLONG)0x400;
							break;
					}
				}
			} else if ( strcmpi(s, "buffer") == 0 && i + 1 < argc && argv[i+1][0] != '-' )
			{
				int			n;

				s = argv[++i];
				n = atoi(s);
				if ( n > 1 && n <= 0x10000 )
					ns.bufferSize = n;
			} else
			{
				fprintf(stderr, "parameter error %s\n", argv[i]);
				return 1;
			}
		} else
		{
			fprintf(stderr, "parameter error %s\n", argv[i]);
			return 1;
		}
	}
	if ( (LONGLONG)ns.bufferSize > ns.transferSize )
		ns.bufferSize = (unsigned)ns.transferSize;
	switch ( ns.mode )
	{
		case NetSpeed::MODE_RECV:
			fprintf(stderr,"Recv mode\n");
			break;

		case NetSpeed::MODE_SEND:
			fprintf(stderr,"Send mode\n");
			if ( ns.host == NULL || ns.host[0] == '\0' )
			{
				fprintf(stderr,"host error!\n");
				return 1;
			}
			break;
	}
	fprintf(stderr, "thread %d, buffer size %u bytes, transfer size %u K bytes\n",
	  ns.threadNum,
	  ns.bufferSize,
	  (unsigned)(ns.transferSize/(LONGLONG)0x400) );
	if ( ns.Run() < 0 )
		return 1;

	return 0;
}


